--  Haskell implementation of Pi by Quadrature
--
--  Copyright © 2009 Russel Winder

module Main where

import Data.Time.Clock ( getCurrentTime , diffUTCTime )

--  Tail recursive functions are not necessarily tail recursive.  This one definitely is not. It is all to
--  do with laziness and thunks.

piIter :: Integer -> Double -> Double -> Double
piIter 0 delta accumulator = 4.0 * accumulator * delta
piIter n delta accumulator = piIter ( n - 1 ) delta ( accumulator + 1.0 / ( 1.0 + x * x )  )
    where
      x = ( ( fromIntegral n ) - 0.5 ) * delta

--  Force strict evaluation using `seq` to ensure a tail recursive function.

strictPiIter :: Integer -> Double -> Double -> Double
strictPiIter 0 delta accumulator = 4.0 * accumulator * delta
strictPiIter n delta accumulator =
    let
        nMinus1 = n -1
        x = ( ( fromIntegral n ) - 0.5 ) * delta
        value = accumulator + 1.0 / ( 1.0 + x * x )
    in
      nMinus1 `seq` x `seq` value `seq` strictPiIter nMinus1 delta value
 
--  Using the standard stack size of 8M can only do n = 1000000 with piIter because even though it appears
--  to be tail recursive it isn't due to the laziness and thunking which prohibits stack squashing.

--  strictPiIter doesn't use much stack but it does take a long time. 

piQuad :: Integer -> Double
piQuad n = strictPiIter n ( 1.0 / ( fromIntegral n ) ) 0.0

main :: IO ( )
main = do
  let n = 100000000 -- 10 times fewer for testing purposes.
  startTime <- getCurrentTime
  print ( "==== Haskell Sequential pi = " ++ show ( piQuad n ) )
  print ( "==== Haskell Sequential iteration count = " ++ show n )
  endTime <- getCurrentTime
  print ( "==== Haskell Sequential elapse time = " ++ show ( diffUTCTime endTime startTime ) )
